"""
Snowflake Grant API.

The Snowflake Grant API is a REST API that you can use to show or manage privileges that have been provided to users and roles in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

import json
import logging
import os
import re
import tempfile

from concurrent.futures import Future
from datetime import date, datetime
from typing import Iterable, Literal, Optional, Union, overload

from dateutil.parser import parse
from pydantic import Field, StrictStr, validate_call
from typing_extensions import Annotated

import snowflake.core._http_requests

from snowflake.core._generated import rest
from snowflake.core._generated.api_client import ApiClient
from snowflake.core._internal.utils import is_running_inside_stored_procedure
from snowflake.core.exceptions import (  # noqa: F401
    _APITypeError,
    _APIValueError,
)
from snowflake.core.grant._generated import models
from snowflake.core.grant._generated.models.grant import Grant
from snowflake.core.grant._generated.models.success_response import SuccessResponse


logger = logging.getLogger(__name__)


class GrantApi:
    """NOTE: This class is auto generated by OpenAPI Generator.

    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, root, resource_class, sproc_client):
        self._root = root
        self._resource_name = "grant"
        self._resource_class = resource_class

        if is_running_inside_stored_procedure():
            self._api_client = sproc_client
        else:
            self._api_client = ApiClient.get_default()

    @property
    def api_client(self):
        return self._api_client

    def deserialize(self, response, response_type):
        """Deserializes response into an object.

        :param response: RESTResponse object to be deserialized.
        :param response_type: class literal for
            deserialized object, or string of class name.

        :return: deserialized object.
        """
        # handle file downloading
        # save response body into a tmp file and return the instance
        if response_type == "file":
            return self.__deserialize_file(response)

        # fetch data from response object
        try:
            data = json.loads(response.data)
        except ValueError:
            data = response.data

        return self.__deserialize(data, response_type)

    def __deserialize(self, data, klass):
        """Deserializes dict, list, str into an object.

        :param data: dict, list or str.
        :param klass: class literal, or string of class name.

        :return: object.
        """
        if data is None:
            return None

        if isinstance(klass, str):
            if klass.startswith("Iterable["):
                sub_kls = re.match(r"Iterable\[(.*)]", klass).group(1)
                return [self.__deserialize(sub_data, sub_kls) for sub_data in data]

            if klass.startswith("Dict["):
                sub_kls = re.match(r"Dict\[([^,]*), (.*)]", klass).group(2)
                return {k: self.__deserialize(v, sub_kls) for k, v in data.items()}

            # convert str to class
            if klass in snowflake.core._http_requests.NATIVE_TYPES_MAPPING:
                klass = snowflake.core._http_requests.NATIVE_TYPES_MAPPING[klass]
            else:
                klass = getattr(models, klass)

        if klass in snowflake.core._http_requests.PRIMITIVE_TYPES:
            return self.__deserialize_primitive(data, klass)
        elif klass == object:  # noqa: E721
            return self.__deserialize_object(data)
        elif klass == date:
            return self.__deserialize_date(data)
        elif klass == datetime:
            return self.__deserialize_datetime(data)
        else:
            return self.__deserialize_model(data, klass)

    def __deserialize_file(self, response):
        """Deserializes body to file.

        Saves response body into a file in a temporary folder,
        using the filename from the `Content-Disposition` header if provided.

        :param response:  RESTResponse.
        :return: file path.
        """
        fd, path = tempfile.mkstemp(dir=self.api_client.configuration.temp_folder_path)
        os.close(fd)
        os.remove(path)

        content_disposition = response.getheader("Content-Disposition")
        if content_disposition:
            filename = re.search(r'filename=[\'"]?([^\'"\s]+)[\'"]?', content_disposition).group(1)
            path = os.path.join(os.path.dirname(path), filename)

        with open(path, "wb") as f:
            f.write(response.data)

        return path

    def __deserialize_primitive(self, data, klass):
        """Deserializes string to primitive type.

        :param data: str.
        :param klass: class literal.

        :return: int, long, float, str, bool.
        """
        try:
            return klass(data)
        except UnicodeEncodeError:
            return str(data)
        except TypeError:
            return data

    def __deserialize_object(self, value):
        """Return an original value.

        :return: object.
        """
        return value

    def __deserialize_date(self, string):
        """Deserializes string to date.

        :param string: str.
        :return: date.
        """
        try:
            return parse(string).date()
        except ImportError:
            return string
        except ValueError:
            raise rest.APIError(status=0, reason=f"Failed to parse `{string}` as date object")

    def __deserialize_datetime(self, string):
        """Deserializes string to datetime.

        The string should be in iso8601 datetime format.

        :param string: str.
        :return: datetime.
        """
        try:
            return parse(string)
        except ImportError:
            return string
        except ValueError:
            raise rest.APIError(status=0, reason=(f"Failed to parse `{string}` as datetime object"))

    def __deserialize_model(self, data, klass):
        """Deserializes list or dict to model.

        :param data: dict, list.
        :param klass: class literal.
        :return: model object.
        """
        return klass.from_dict(data)

    @overload
    def grant_group_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        grant: Grant,
        async_req: Literal[False] = False,
        **kwargs,
    ) -> SuccessResponse: ...

    @overload
    def grant_group_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        grant: Grant,
        async_req: Literal[True] = True,
        **kwargs,
    ) -> Future[SuccessResponse]: ...

    @overload
    def grant_group_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        grant: Grant,
        async_req: bool = False,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]: ...

    @validate_call
    def grant_group_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        grant: Grant,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]:
        r"""Grant the specified privilege(s) on all/future (as specified by bulkGrantType) securables of this type in a given scope to the named grantee.  # noqa: E501

        Endpoint to indicate that the privileges listed in the request body should be granted to all securables of this type in the given scope.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.grant_group_privilege(
        ...     grantee_type,
        ...     grantee_name,
        ...     bulk_grant_type,
        ...     securable_type_plural,
        ...     scope_type,
        ...     scope_name,
        ...     grant,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param bulk_grant_type: String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type (required)
        :type bulk_grant_type: str
        :param securable_type_plural: String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either \"schemas\" or any plural object type that can nest under a schema such as \"tables\" (required)
        :type securable_type_plural: str
        :param scope_type: String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA (required)
        :type scope_type: str
        :param scope_name: String that specifies the name of resource that is the scope of an ALL/FUTURE privilege (required)
        :type scope_name: str
        :param grant: (required)
        :type grant: Grant
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: Union[SuccessResponse, Future[SuccessResponse]]
        """
        kwargs["_return_http_data_only"] = True
        return self.grant_group_privilege_with_http_info(
            grantee_type, grantee_name, bulk_grant_type, securable_type_plural, scope_type, scope_name, grant, **kwargs
        )

    @validate_call
    def grant_group_privilege_with_http_info(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        grant: Grant,
        **kwargs,
    ):
        r"""Grant the specified privilege(s) on all/future (as specified by bulkGrantType) securables of this type in a given scope to the named grantee.  # noqa: E501

        Endpoint to indicate that the privileges listed in the request body should be granted to all securables of this type in the given scope.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.grant_group_privilege_with_http_info(
        ...     grantee_type,
        ...     grantee_name,
        ...     bulk_grant_type,
        ...     securable_type_plural,
        ...     scope_type,
        ...     scope_name,
        ...     grant,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param bulk_grant_type: String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type (required)
        :type bulk_grant_type: str
        :param securable_type_plural: String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either \"schemas\" or any plural object type that can nest under a schema such as \"tables\" (required)
        :type securable_type_plural: str
        :param scope_type: String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA (required)
        :type scope_type: str
        :param scope_name: String that specifies the name of resource that is the scope of an ALL/FUTURE privilege (required)
        :type scope_name: str
        :param grant: (required)
        :type grant: Grant
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: tuple(Union[SuccessResponse, Future[SuccessResponse]], status_code(int), headers(HTTPHeaderDict))
        """
        _params = locals()

        _all_params = [
            "grantee_type",
            "grantee_name",
            "bulk_grant_type",
            "securable_type_plural",
            "scope_type",
            "scope_name",
            "grant",
        ]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise _APITypeError(f"Got an unexpected keyword argument '{_key}' to method grant_group_privilege")
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        if _params["grantee_type"]:
            _path_params["granteeType"] = _params["grantee_type"]

        if _params["grantee_name"]:
            _path_params["granteeName"] = _params["grantee_name"]

        if _params["bulk_grant_type"]:
            _path_params["bulkGrantType"] = _params["bulk_grant_type"]

        if _params["securable_type_plural"]:
            _path_params["securableTypePlural"] = _params["securable_type_plural"]

        if _params["scope_type"]:
            _path_params["scopeType"] = _params["scope_type"]

        if _params["scope_name"]:
            _path_params["scopeName"] = _params["scope_name"]

        # process the query parameters
        _query_params = []

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        if _params["grant"] is not None:
            _body_params = _params["grant"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(["application/json"])

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type", self.api_client.select_header_content_type(["application/json"])
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["ExternalOAuth", "KeyPair", "ProgrammaticAccessToken", "SnowflakeOAuth"]

        _response_types_map = {
            "200": "SuccessResponse",
            "202": "SuccessAcceptedResponse",
            "400": "ErrorResponse",
            "401": "ErrorResponse",
            "403": "ErrorResponse",
            "404": "ErrorResponse",
            "405": "ErrorResponse",
            "408": "ErrorResponse",
            "409": "ErrorResponse",
            "410": "ErrorResponse",
            "429": "ErrorResponse",
            "500": "ErrorResponse",
            "503": "ErrorResponse",
            "504": "ErrorResponse",
        }

        return self.api_client.call_api(
            self._root,
            "/api/v2/grants/{granteeType}/{granteeName}/{bulkGrantType}/{securableTypePlural}/{scopeType}/{scopeName}/privileges",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
            _deserialize_response_fn=self.deserialize,
        )

    @overload
    def grant_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        grant: Grant,
        async_req: Literal[False] = False,
        **kwargs,
    ) -> SuccessResponse: ...

    @overload
    def grant_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        grant: Grant,
        async_req: Literal[True] = True,
        **kwargs,
    ) -> Future[SuccessResponse]: ...

    @overload
    def grant_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        grant: Grant,
        async_req: bool = False,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]: ...

    @validate_call
    def grant_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        grant: Grant,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]:
        r"""Grant the specified privilege(s) on the named securable to the named grantee.  # noqa: E501

        Endpoint to indicate that the privileges listed in the request body should be granted.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.grant_privilege(
        ...     grantee_type, grantee_name, securable_type, securable_name, grant, async_req=True
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param securable_type: String that specifies the type of resource that is being secured by a privilege. (required)
        :type securable_type: str
        :param securable_name: String that specifies the name of resource that is being secured by a privilege. (required)
        :type securable_name: str
        :param grant: (required)
        :type grant: Grant
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: Union[SuccessResponse, Future[SuccessResponse]]
        """
        kwargs["_return_http_data_only"] = True
        return self.grant_privilege_with_http_info(
            grantee_type, grantee_name, securable_type, securable_name, grant, **kwargs
        )

    @validate_call
    def grant_privilege_with_http_info(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        grant: Grant,
        **kwargs,
    ):
        r"""Grant the specified privilege(s) on the named securable to the named grantee.  # noqa: E501

        Endpoint to indicate that the privileges listed in the request body should be granted.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.grant_privilege_with_http_info(
        ...     grantee_type, grantee_name, securable_type, securable_name, grant, async_req=True
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param securable_type: String that specifies the type of resource that is being secured by a privilege. (required)
        :type securable_type: str
        :param securable_name: String that specifies the name of resource that is being secured by a privilege. (required)
        :type securable_name: str
        :param grant: (required)
        :type grant: Grant
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: tuple(Union[SuccessResponse, Future[SuccessResponse]], status_code(int), headers(HTTPHeaderDict))
        """
        _params = locals()

        _all_params = ["grantee_type", "grantee_name", "securable_type", "securable_name", "grant"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise _APITypeError(f"Got an unexpected keyword argument '{_key}' to method grant_privilege")
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        if _params["grantee_type"]:
            _path_params["granteeType"] = _params["grantee_type"]

        if _params["grantee_name"]:
            _path_params["granteeName"] = _params["grantee_name"]

        if _params["securable_type"]:
            _path_params["securableType"] = _params["securable_type"]

        if _params["securable_name"]:
            _path_params["securableName"] = _params["securable_name"]

        # process the query parameters
        _query_params = []

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        if _params["grant"] is not None:
            _body_params = _params["grant"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(["application/json"])

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type", self.api_client.select_header_content_type(["application/json"])
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["ExternalOAuth", "KeyPair", "ProgrammaticAccessToken", "SnowflakeOAuth"]

        _response_types_map = {
            "200": "SuccessResponse",
            "202": "SuccessAcceptedResponse",
            "400": "ErrorResponse",
            "401": "ErrorResponse",
            "403": "ErrorResponse",
            "404": "ErrorResponse",
            "405": "ErrorResponse",
            "408": "ErrorResponse",
            "409": "ErrorResponse",
            "410": "ErrorResponse",
            "429": "ErrorResponse",
            "500": "ErrorResponse",
            "503": "ErrorResponse",
            "504": "ErrorResponse",
        }

        return self.api_client.call_api(
            self._root,
            "/api/v2/grants/{granteeType}/{granteeName}/{securableType}/{securableName}/privileges",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
            _deserialize_response_fn=self.deserialize,
        )

    @overload
    def list_grants_to(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        show_limit: Annotated[
            Optional[Annotated[int, Field(le=10000, strict=True, ge=1)]],
            Field(description="Parameter to limit the maximum number of rows returned by a command."),
        ] = None,
        async_req: Literal[False] = False,
        **kwargs,
    ) -> Iterable[Grant]: ...

    @overload
    def list_grants_to(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        show_limit: Annotated[
            Optional[Annotated[int, Field(le=10000, strict=True, ge=1)]],
            Field(description="Parameter to limit the maximum number of rows returned by a command."),
        ] = None,
        async_req: Literal[True] = True,
        **kwargs,
    ) -> Future[Iterable[Grant]]: ...

    @overload
    def list_grants_to(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        show_limit: Annotated[
            Optional[Annotated[int, Field(le=10000, strict=True, ge=1)]],
            Field(description="Parameter to limit the maximum number of rows returned by a command."),
        ] = None,
        async_req: bool = False,
        **kwargs,
    ) -> Union[Iterable[Grant], Future[Iterable[Grant]]]: ...

    @validate_call
    def list_grants_to(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        show_limit: Annotated[
            Optional[Annotated[int, Field(le=10000, strict=True, ge=1)]],
            Field(description="Parameter to limit the maximum number of rows returned by a command."),
        ] = None,
        **kwargs,
    ) -> Union[Iterable[Grant], Future[Iterable[Grant]]]:
        r"""List of privileges associated with this grantee type and name  # noqa: E501

        List the roles and privileges granted to the specified grantee using the output of SHOW GRANTS TO  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.list_grants_to(grantee_type, grantee_name, show_limit, async_req=True)
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param show_limit: Parameter to limit the maximum number of rows returned by a command.
        :type show_limit: int
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: Union[Iterable[Grant], Future[Iterable[Grant]]]
        """
        kwargs["_return_http_data_only"] = True
        return self.list_grants_to_with_http_info(grantee_type, grantee_name, show_limit, **kwargs)

    @validate_call
    def list_grants_to_with_http_info(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        show_limit: Annotated[
            Optional[Annotated[int, Field(le=10000, strict=True, ge=1)]],
            Field(description="Parameter to limit the maximum number of rows returned by a command."),
        ] = None,
        **kwargs,
    ):
        r"""List of privileges associated with this grantee type and name  # noqa: E501

        List the roles and privileges granted to the specified grantee using the output of SHOW GRANTS TO  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.list_grants_to_with_http_info(grantee_type, grantee_name, show_limit, async_req=True)
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param show_limit: Parameter to limit the maximum number of rows returned by a command.
        :type show_limit: int
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: tuple(Union[Iterable[Grant], Future[Iterable[Grant]]], status_code(int), headers(HTTPHeaderDict))
        """
        _params = locals()

        _all_params = ["grantee_type", "grantee_name", "show_limit"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise _APITypeError(f"Got an unexpected keyword argument '{_key}' to method list_grants_to")
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        if _params["grantee_type"]:
            _path_params["granteeType"] = _params["grantee_type"]

        if _params["grantee_name"]:
            _path_params["granteeName"] = _params["grantee_name"]

        # process the query parameters
        _query_params = []

        if _params.get("show_limit") is not None:
            _query_params.append(("showLimit", _params["show_limit"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(["application/json"])

        # authentication setting
        _auth_settings = ["ExternalOAuth", "KeyPair", "ProgrammaticAccessToken", "SnowflakeOAuth"]

        _response_types_map = {
            "200": "Iterable[Grant]",
            "202": "SuccessAcceptedResponse",
            "400": "ErrorResponse",
            "401": "ErrorResponse",
            "403": "ErrorResponse",
            "404": "ErrorResponse",
            "405": "ErrorResponse",
            "408": "ErrorResponse",
            "409": "ErrorResponse",
            "410": "ErrorResponse",
            "429": "ErrorResponse",
            "500": "ErrorResponse",
            "503": "ErrorResponse",
            "504": "ErrorResponse",
        }

        return self.api_client.call_api(
            self._root,
            "/api/v2/grants/{granteeType}/{granteeName}",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
            _deserialize_response_fn=self.deserialize,
        )

    @overload
    def revoke_group_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: Literal[False] = False,
        **kwargs,
    ) -> SuccessResponse: ...

    @overload
    def revoke_group_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: Literal[True] = True,
        **kwargs,
    ) -> Future[SuccessResponse]: ...

    @overload
    def revoke_group_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: bool = False,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]: ...

    @validate_call
    def revoke_group_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]:
        r"""Revoke the specified privilege on the specified all/future securable in the given scope from the named grantee.  # noqa: E501

        Endpoint to indicate that the privilege listed on the group securable in the given scope should be revoked.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.revoke_group_privilege(
        ...     grantee_type,
        ...     grantee_name,
        ...     bulk_grant_type,
        ...     securable_type_plural,
        ...     scope_type,
        ...     scope_name,
        ...     privilege,
        ...     delete_mode,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param bulk_grant_type: String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type (required)
        :type bulk_grant_type: str
        :param securable_type_plural: String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either \"schemas\" or any plural object type that can nest under a schema such as \"tables\" (required)
        :type securable_type_plural: str
        :param scope_type: String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA (required)
        :type scope_type: str
        :param scope_name: String that specifies the name of resource that is the scope of an ALL/FUTURE privilege (required)
        :type scope_name: str
        :param privilege: String that specifies a privilege to be revoked (required)
        :type privilege: str
        :param delete_mode: If \"cascade\", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are \"restrict\" or \"cascade\".
        :type delete_mode: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: Union[SuccessResponse, Future[SuccessResponse]]
        """
        kwargs["_return_http_data_only"] = True
        return self.revoke_group_privilege_with_http_info(
            grantee_type,
            grantee_name,
            bulk_grant_type,
            securable_type_plural,
            scope_type,
            scope_name,
            privilege,
            delete_mode,
            **kwargs,
        )

    @validate_call
    def revoke_group_privilege_with_http_info(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        **kwargs,
    ):
        r"""Revoke the specified privilege on the specified all/future securable in the given scope from the named grantee.  # noqa: E501

        Endpoint to indicate that the privilege listed on the group securable in the given scope should be revoked.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.revoke_group_privilege_with_http_info(
        ...     grantee_type,
        ...     grantee_name,
        ...     bulk_grant_type,
        ...     securable_type_plural,
        ...     scope_type,
        ...     scope_name,
        ...     privilege,
        ...     delete_mode,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param bulk_grant_type: String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type (required)
        :type bulk_grant_type: str
        :param securable_type_plural: String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either \"schemas\" or any plural object type that can nest under a schema such as \"tables\" (required)
        :type securable_type_plural: str
        :param scope_type: String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA (required)
        :type scope_type: str
        :param scope_name: String that specifies the name of resource that is the scope of an ALL/FUTURE privilege (required)
        :type scope_name: str
        :param privilege: String that specifies a privilege to be revoked (required)
        :type privilege: str
        :param delete_mode: If \"cascade\", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are \"restrict\" or \"cascade\".
        :type delete_mode: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: tuple(Union[SuccessResponse, Future[SuccessResponse]], status_code(int), headers(HTTPHeaderDict))
        """
        _params = locals()

        _all_params = [
            "grantee_type",
            "grantee_name",
            "bulk_grant_type",
            "securable_type_plural",
            "scope_type",
            "scope_name",
            "privilege",
            "delete_mode",
        ]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise _APITypeError(f"Got an unexpected keyword argument '{_key}' to method revoke_group_privilege")
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        if _params["grantee_type"]:
            _path_params["granteeType"] = _params["grantee_type"]

        if _params["grantee_name"]:
            _path_params["granteeName"] = _params["grantee_name"]

        if _params["bulk_grant_type"]:
            _path_params["bulkGrantType"] = _params["bulk_grant_type"]

        if _params["securable_type_plural"]:
            _path_params["securableTypePlural"] = _params["securable_type_plural"]

        if _params["scope_type"]:
            _path_params["scopeType"] = _params["scope_type"]

        if _params["scope_name"]:
            _path_params["scopeName"] = _params["scope_name"]

        if _params["privilege"]:
            _path_params["privilege"] = _params["privilege"]

        # process the query parameters
        _query_params = []

        if _params.get("delete_mode") is not None:
            _query_params.append(("deleteMode", _params["delete_mode"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(["application/json"])

        # authentication setting
        _auth_settings = ["ExternalOAuth", "KeyPair", "ProgrammaticAccessToken", "SnowflakeOAuth"]

        _response_types_map = {
            "200": "SuccessResponse",
            "202": "SuccessAcceptedResponse",
            "400": "ErrorResponse",
            "401": "ErrorResponse",
            "403": "ErrorResponse",
            "404": "ErrorResponse",
            "405": "ErrorResponse",
            "408": "ErrorResponse",
            "409": "ErrorResponse",
            "410": "ErrorResponse",
            "429": "ErrorResponse",
            "500": "ErrorResponse",
            "503": "ErrorResponse",
            "504": "ErrorResponse",
        }

        return self.api_client.call_api(
            self._root,
            "/api/v2/grants/{granteeType}/{granteeName}/{bulkGrantType}/{securableTypePlural}/{scopeType}/{scopeName}/privileges/{privilege}",
            "DELETE",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
            _deserialize_response_fn=self.deserialize,
        )

    @overload
    def revoke_group_privilege_grant_option(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: Literal[False] = False,
        **kwargs,
    ) -> SuccessResponse: ...

    @overload
    def revoke_group_privilege_grant_option(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: Literal[True] = True,
        **kwargs,
    ) -> Future[SuccessResponse]: ...

    @overload
    def revoke_group_privilege_grant_option(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: bool = False,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]: ...

    @validate_call
    def revoke_group_privilege_grant_option(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]:
        r"""Revoke the grant option for the specified privilege on the specified all/future securable in the given scope from the named grantee.  # noqa: E501

        Endpoint to indicate that the grant option for the privilege listed on the group securable in the given scope should be revoked.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.revoke_group_privilege_grant_option(
        ...     grantee_type,
        ...     grantee_name,
        ...     bulk_grant_type,
        ...     securable_type_plural,
        ...     scope_type,
        ...     scope_name,
        ...     privilege,
        ...     delete_mode,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param bulk_grant_type: String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type (required)
        :type bulk_grant_type: str
        :param securable_type_plural: String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either \"schemas\" or any plural object type that can nest under a schema such as \"tables\" (required)
        :type securable_type_plural: str
        :param scope_type: String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA (required)
        :type scope_type: str
        :param scope_name: String that specifies the name of resource that is the scope of an ALL/FUTURE privilege (required)
        :type scope_name: str
        :param privilege: String that specifies a privilege to be revoked (required)
        :type privilege: str
        :param delete_mode: If \"cascade\", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are \"restrict\" or \"cascade\".
        :type delete_mode: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: Union[SuccessResponse, Future[SuccessResponse]]
        """
        kwargs["_return_http_data_only"] = True
        return self.revoke_group_privilege_grant_option_with_http_info(
            grantee_type,
            grantee_name,
            bulk_grant_type,
            securable_type_plural,
            scope_type,
            scope_name,
            privilege,
            delete_mode,
            **kwargs,
        )

    @validate_call
    def revoke_group_privilege_grant_option_with_http_info(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        bulk_grant_type: Annotated[
            StrictStr,
            Field(
                description="String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type"
            ),
        ],
        securable_type_plural: Annotated[
            StrictStr,
            Field(
                description='String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either "schemas" or any plural object type that can nest under a schema such as "tables"'
            ),
        ],
        scope_type: Annotated[
            StrictStr,
            Field(
                description="String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA"
            ),
        ],
        scope_name: Annotated[
            StrictStr,
            Field(
                description="String that specifies the name of resource that is the scope of an ALL/FUTURE privilege"
            ),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        **kwargs,
    ):
        r"""Revoke the grant option for the specified privilege on the specified all/future securable in the given scope from the named grantee.  # noqa: E501

        Endpoint to indicate that the grant option for the privilege listed on the group securable in the given scope should be revoked.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.revoke_group_privilege_grant_option_with_http_info(
        ...     grantee_type,
        ...     grantee_name,
        ...     bulk_grant_type,
        ...     securable_type_plural,
        ...     scope_type,
        ...     scope_name,
        ...     privilege,
        ...     delete_mode,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param bulk_grant_type: String that species whether this group privilege should be on ALL or FUTURE resources of the specified plural type (required)
        :type bulk_grant_type: str
        :param securable_type_plural: String that specifies the plural of the type of resource that is being secured by an ALL/FUTURE privilege. Must be either \"schemas\" or any plural object type that can nest under a schema such as \"tables\" (required)
        :type securable_type_plural: str
        :param scope_type: String that specifies the type of resource that is the scope of an ALL/FUTURE privilege. Can only be DATABASE or SCHEMA (required)
        :type scope_type: str
        :param scope_name: String that specifies the name of resource that is the scope of an ALL/FUTURE privilege (required)
        :type scope_name: str
        :param privilege: String that specifies a privilege to be revoked (required)
        :type privilege: str
        :param delete_mode: If \"cascade\", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are \"restrict\" or \"cascade\".
        :type delete_mode: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: tuple(Union[SuccessResponse, Future[SuccessResponse]], status_code(int), headers(HTTPHeaderDict))
        """
        _params = locals()

        _all_params = [
            "grantee_type",
            "grantee_name",
            "bulk_grant_type",
            "securable_type_plural",
            "scope_type",
            "scope_name",
            "privilege",
            "delete_mode",
        ]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise _APITypeError(
                    f"Got an unexpected keyword argument '{_key}' to method revoke_group_privilege_grant_option"
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        if _params["grantee_type"]:
            _path_params["granteeType"] = _params["grantee_type"]

        if _params["grantee_name"]:
            _path_params["granteeName"] = _params["grantee_name"]

        if _params["bulk_grant_type"]:
            _path_params["bulkGrantType"] = _params["bulk_grant_type"]

        if _params["securable_type_plural"]:
            _path_params["securableTypePlural"] = _params["securable_type_plural"]

        if _params["scope_type"]:
            _path_params["scopeType"] = _params["scope_type"]

        if _params["scope_name"]:
            _path_params["scopeName"] = _params["scope_name"]

        if _params["privilege"]:
            _path_params["privilege"] = _params["privilege"]

        # process the query parameters
        _query_params = []

        if _params.get("delete_mode") is not None:
            _query_params.append(("deleteMode", _params["delete_mode"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(["application/json"])

        # authentication setting
        _auth_settings = ["ExternalOAuth", "KeyPair", "ProgrammaticAccessToken", "SnowflakeOAuth"]

        _response_types_map = {
            "200": "SuccessResponse",
            "202": "SuccessAcceptedResponse",
            "400": "ErrorResponse",
            "401": "ErrorResponse",
            "403": "ErrorResponse",
            "404": "ErrorResponse",
            "405": "ErrorResponse",
            "408": "ErrorResponse",
            "409": "ErrorResponse",
            "410": "ErrorResponse",
            "429": "ErrorResponse",
            "500": "ErrorResponse",
            "503": "ErrorResponse",
            "504": "ErrorResponse",
        }

        return self.api_client.call_api(
            self._root,
            "/api/v2/grants/{granteeType}/{granteeName}/{bulkGrantType}/{securableTypePlural}/{scopeType}/{scopeName}/privileges/{privilege}/grant-option",
            "DELETE",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
            _deserialize_response_fn=self.deserialize,
        )

    @overload
    def revoke_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: Literal[False] = False,
        **kwargs,
    ) -> SuccessResponse: ...

    @overload
    def revoke_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: Literal[True] = True,
        **kwargs,
    ) -> Future[SuccessResponse]: ...

    @overload
    def revoke_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: bool = False,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]: ...

    @validate_call
    def revoke_privilege(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]:
        r"""Revoke the specified privilege on the named securable from the named grantee.  # noqa: E501

        Endpoint to indicate that the privilege listed in the path should be revoked.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.revoke_privilege(
        ...     grantee_type,
        ...     grantee_name,
        ...     securable_type,
        ...     securable_name,
        ...     privilege,
        ...     delete_mode,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param securable_type: String that specifies the type of resource that is being secured by a privilege. (required)
        :type securable_type: str
        :param securable_name: String that specifies the name of resource that is being secured by a privilege. (required)
        :type securable_name: str
        :param privilege: String that specifies a privilege to be revoked (required)
        :type privilege: str
        :param delete_mode: If \"cascade\", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are \"restrict\" or \"cascade\".
        :type delete_mode: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: Union[SuccessResponse, Future[SuccessResponse]]
        """
        kwargs["_return_http_data_only"] = True
        return self.revoke_privilege_with_http_info(
            grantee_type, grantee_name, securable_type, securable_name, privilege, delete_mode, **kwargs
        )

    @validate_call
    def revoke_privilege_with_http_info(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        **kwargs,
    ):
        r"""Revoke the specified privilege on the named securable from the named grantee.  # noqa: E501

        Endpoint to indicate that the privilege listed in the path should be revoked.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.revoke_privilege_with_http_info(
        ...     grantee_type,
        ...     grantee_name,
        ...     securable_type,
        ...     securable_name,
        ...     privilege,
        ...     delete_mode,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param securable_type: String that specifies the type of resource that is being secured by a privilege. (required)
        :type securable_type: str
        :param securable_name: String that specifies the name of resource that is being secured by a privilege. (required)
        :type securable_name: str
        :param privilege: String that specifies a privilege to be revoked (required)
        :type privilege: str
        :param delete_mode: If \"cascade\", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are \"restrict\" or \"cascade\".
        :type delete_mode: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: tuple(Union[SuccessResponse, Future[SuccessResponse]], status_code(int), headers(HTTPHeaderDict))
        """
        _params = locals()

        _all_params = ["grantee_type", "grantee_name", "securable_type", "securable_name", "privilege", "delete_mode"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise _APITypeError(f"Got an unexpected keyword argument '{_key}' to method revoke_privilege")
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        if _params["grantee_type"]:
            _path_params["granteeType"] = _params["grantee_type"]

        if _params["grantee_name"]:
            _path_params["granteeName"] = _params["grantee_name"]

        if _params["securable_type"]:
            _path_params["securableType"] = _params["securable_type"]

        if _params["securable_name"]:
            _path_params["securableName"] = _params["securable_name"]

        if _params["privilege"]:
            _path_params["privilege"] = _params["privilege"]

        # process the query parameters
        _query_params = []

        if _params.get("delete_mode") is not None:
            _query_params.append(("deleteMode", _params["delete_mode"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(["application/json"])

        # authentication setting
        _auth_settings = ["ExternalOAuth", "KeyPair", "ProgrammaticAccessToken", "SnowflakeOAuth"]

        _response_types_map = {
            "200": "SuccessResponse",
            "202": "SuccessAcceptedResponse",
            "400": "ErrorResponse",
            "401": "ErrorResponse",
            "403": "ErrorResponse",
            "404": "ErrorResponse",
            "405": "ErrorResponse",
            "408": "ErrorResponse",
            "409": "ErrorResponse",
            "410": "ErrorResponse",
            "429": "ErrorResponse",
            "500": "ErrorResponse",
            "503": "ErrorResponse",
            "504": "ErrorResponse",
        }

        return self.api_client.call_api(
            self._root,
            "/api/v2/grants/{granteeType}/{granteeName}/{securableType}/{securableName}/privileges/{privilege}",
            "DELETE",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
            _deserialize_response_fn=self.deserialize,
        )

    @overload
    def revoke_privilege_grant_option(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: Literal[False] = False,
        **kwargs,
    ) -> SuccessResponse: ...

    @overload
    def revoke_privilege_grant_option(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: Literal[True] = True,
        **kwargs,
    ) -> Future[SuccessResponse]: ...

    @overload
    def revoke_privilege_grant_option(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        async_req: bool = False,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]: ...

    @validate_call
    def revoke_privilege_grant_option(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        **kwargs,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]:
        r"""Revoke the grant option for the specified privilege on the named securable from the named grantee.  # noqa: E501

        Endpoint to indicate that the grant option for the privilege listed in the path should be revoked.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.revoke_privilege_grant_option(
        ...     grantee_type,
        ...     grantee_name,
        ...     securable_type,
        ...     securable_name,
        ...     privilege,
        ...     delete_mode,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param securable_type: String that specifies the type of resource that is being secured by a privilege. (required)
        :type securable_type: str
        :param securable_name: String that specifies the name of resource that is being secured by a privilege. (required)
        :type securable_name: str
        :param privilege: String that specifies a privilege to be revoked (required)
        :type privilege: str
        :param delete_mode: If \"cascade\", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are \"restrict\" or \"cascade\".
        :type delete_mode: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: Union[SuccessResponse, Future[SuccessResponse]]
        """
        kwargs["_return_http_data_only"] = True
        return self.revoke_privilege_grant_option_with_http_info(
            grantee_type, grantee_name, securable_type, securable_name, privilege, delete_mode, **kwargs
        )

    @validate_call
    def revoke_privilege_grant_option_with_http_info(
        self,
        grantee_type: Annotated[
            StrictStr, Field(description="String that specifies the type of resource that is the privilege grantee.")
        ],
        grantee_name: Annotated[
            StrictStr, Field(description="String that specifies the name of the privilege grantee.")
        ],
        securable_type: Annotated[
            StrictStr,
            Field(description="String that specifies the type of resource that is being secured by a privilege."),
        ],
        securable_name: Annotated[
            StrictStr,
            Field(description="String that specifies the name of resource that is being secured by a privilege."),
        ],
        privilege: Annotated[StrictStr, Field(description="String that specifies a privilege to be revoked")],
        delete_mode: Annotated[
            Optional[StrictStr],
            Field(
                description='If "cascade", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are "restrict" or "cascade".'
            ),
        ] = None,
        **kwargs,
    ):
        r"""Revoke the grant option for the specified privilege on the named securable from the named grantee.  # noqa: E501

        Endpoint to indicate that the grant option for the privilege listed in the path should be revoked.  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> future = api.revoke_privilege_grant_option_with_http_info(
        ...     grantee_type,
        ...     grantee_name,
        ...     securable_type,
        ...     securable_name,
        ...     privilege,
        ...     delete_mode,
        ...     async_req=True,
        ... )
        >>> result = future.result()
        :param grantee_type: String that specifies the type of resource that is the privilege grantee. (required)
        :type grantee_type: str
        :param grantee_name: String that specifies the name of the privilege grantee. (required)
        :type grantee_name: str
        :param securable_type: String that specifies the type of resource that is being secured by a privilege. (required)
        :type securable_type: str
        :param securable_name: String that specifies the name of resource that is being secured by a privilege. (required)
        :type securable_name: str
        :param privilege: String that specifies a privilege to be revoked (required)
        :type privilege: str
        :param delete_mode: If \"cascade\", recursively revoke the grant from sub-grantees to which this privilege was re-granted. Acceptable values are \"restrict\" or \"cascade\".
        :type delete_mode: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns a Future object representing the execution of the method.
        :rtype: tuple(Union[SuccessResponse, Future[SuccessResponse]], status_code(int), headers(HTTPHeaderDict))
        """
        _params = locals()

        _all_params = ["grantee_type", "grantee_name", "securable_type", "securable_name", "privilege", "delete_mode"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise _APITypeError(
                    f"Got an unexpected keyword argument '{_key}' to method revoke_privilege_grant_option"
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        if _params["grantee_type"]:
            _path_params["granteeType"] = _params["grantee_type"]

        if _params["grantee_name"]:
            _path_params["granteeName"] = _params["grantee_name"]

        if _params["securable_type"]:
            _path_params["securableType"] = _params["securable_type"]

        if _params["securable_name"]:
            _path_params["securableName"] = _params["securable_name"]

        if _params["privilege"]:
            _path_params["privilege"] = _params["privilege"]

        # process the query parameters
        _query_params = []

        if _params.get("delete_mode") is not None:
            _query_params.append(("deleteMode", _params["delete_mode"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(["application/json"])

        # authentication setting
        _auth_settings = ["ExternalOAuth", "KeyPair", "ProgrammaticAccessToken", "SnowflakeOAuth"]

        _response_types_map = {
            "200": "SuccessResponse",
            "202": "SuccessAcceptedResponse",
            "400": "ErrorResponse",
            "401": "ErrorResponse",
            "403": "ErrorResponse",
            "404": "ErrorResponse",
            "405": "ErrorResponse",
            "408": "ErrorResponse",
            "409": "ErrorResponse",
            "410": "ErrorResponse",
            "429": "ErrorResponse",
            "500": "ErrorResponse",
            "503": "ErrorResponse",
            "504": "ErrorResponse",
        }

        return self.api_client.call_api(
            self._root,
            "/api/v2/grants/{granteeType}/{granteeName}/{securableType}/{securableName}/privileges/{privilege}/grant-option",
            "DELETE",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
            _deserialize_response_fn=self.deserialize,
        )



import pytest

from tests.utils import random_string

from .utils import create_service_function


TRANSLATE_DATA_TYPE_TO_BIND = {"INT": "FIXED"}

pytestmark = [pytest.mark.skip_gov]


@pytest.mark.usefixtures("qa_mode_enabled")
def test_execute_function(temp_service_for_function, functions, setup_with_connector_execution):
    types = [["REAL"], ["INT"], ["BOOLEAN"], ["REAL", "INT"]]
    inputs = [[12.3], [12], [True], [12, 1]]
    outputs = [12.3, 12, True, 12.0]

    alter_prefix = "alter session "
    with setup_with_connector_execution(
        [
            alter_prefix + "set QA_MODE_MOCK_EXTERNAL_FUNCTION_REMOTE_CALLS = true",
            alter_prefix
            + """set snowservices_mock_server_endpoints =
                  '{"ep1":["mockhost1", "mockhost2"],"end-point-2":["mockhost3"]}';
            """,
        ],
        [
            alter_prefix + "unset QA_MODE_MOCK_EXTERNAL_FUNCTION_REMOTE_CALLS",
            alter_prefix + "unset snowservices_mock_server_endpoints",
        ],
    ):
        for i in range(len(inputs)):
            t = types[i]
            function_name = random_string(5, "test_func_")
            function_name_with_args = f"{function_name}({','.join(t)})"
            endpoint = "end-point-2" if t == "INT" else "ep1"
            try:
                f = create_service_function(function_name, t, t[0], endpoint, temp_service_for_function.name, functions)
                assert f.execute(inputs[i]) == outputs[i]
            finally:
                functions[function_name_with_args].drop()



import copy

import pytest

from snowflake.core import CreateMode
from snowflake.core.exceptions import ConflictError
from tests.integ.utils import random_string

from .conftest import (
    test_network_rule_minimal_template,
    test_network_rule_template,
)


def test_create_and_fetch(network_rules):
    name = random_string(10, "test_network_rule_create_and_fetch_")
    network_rule_handle = network_rules[name]

    try:
        network_rule = copy.deepcopy(test_network_rule_template)
        network_rule.name = name
        network_rules.create(network_rule)

        fetched_rule = network_rule_handle.fetch()

        assert fetched_rule.name.upper() == name.upper()
        assert fetched_rule.type == "HOST_PORT"
        assert fetched_rule.mode.upper() == "EGRESS"
        assert fetched_rule.value_list == ["example.com:443", "api.example.com:443"]
        assert fetched_rule.comment == "Test network rule"
    finally:
        network_rule_handle.drop(if_exists=True)


def test_create_and_fetch_minimal(network_rules):
    name = random_string(10, "test_network_rule_create_and_fetch_")
    network_rule_handle = network_rules[name]

    try:
        network_rule = copy.deepcopy(test_network_rule_minimal_template)
        network_rule.name = name
        network_rules.create(network_rule)

        fetched_rule = network_rule_handle.fetch()

        assert fetched_rule.name.upper() == name.upper()
        assert fetched_rule.type == "IPv4"
        assert fetched_rule.mode.upper() == "INGRESS"
        assert fetched_rule.value_list == []
        assert fetched_rule.comment is None
    finally:
        network_rule_handle.drop(if_exists=True)


def test_create_and_fetch_create_modes(network_rules):
    name = random_string(10, "test_network_rule_create_and_fetch_")
    network_rule_handle = network_rules[name]

    try:
        network_rule = copy.deepcopy(test_network_rule_template)
        network_rule.name = name
        network_rule.comment = "First version"
        network_rules.create(network_rule, mode=CreateMode.error_if_exists)
        assert network_rule_handle.fetch().comment == "First version"

        with pytest.raises(ConflictError):
            network_rules.create(network_rule, mode=CreateMode.error_if_exists)

        network_rule.comment = "Second version"
        network_rules.create(network_rule, mode=CreateMode.or_replace)
        assert network_rule_handle.fetch().comment == "Second version"

        network_rule.comment = "Should not change"
        network_rules.create(network_rule, mode=CreateMode.if_not_exists)
        assert network_rule_handle.fetch().comment == "Second version"
    finally:
        network_rule_handle.drop(if_exists=True)



from unittest import mock

import pytest

from snowflake.core import PollingOperation
from snowflake.core.managed_account import ManagedAccount, ManagedAccountCollection, ManagedAccountResource

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
MANAGED_ACCOUNT = ManagedAccount(name="my_acc", admin_name="admin", admin_password="test")


@pytest.fixture
def managed_accounts(fake_root):
    return ManagedAccountCollection(fake_root)


@pytest.fixture
def managed_account(managed_accounts):
    return managed_accounts["my_acc"]


def test_create_managed_account(fake_root, managed_accounts):
    args = (fake_root, "POST", BASE_URL + "/managed-accounts")
    kwargs = extra_params(
        body={"name": "my_acc", "admin_name": "admin", "admin_password": "test", "account_type": "READER"}
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        ma_res = managed_accounts.create(MANAGED_ACCOUNT)
        assert isinstance(ma_res, ManagedAccountResource)
        assert ma_res.name == "my_acc"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = managed_accounts.create_async(MANAGED_ACCOUNT)
        assert isinstance(op, PollingOperation)
        et_res = op.result()
        assert et_res.name == "my_acc"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_managed_account(fake_root, managed_accounts):
    args = (fake_root, "GET", BASE_URL + "/managed-accounts")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        managed_accounts.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = managed_accounts.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_managed_account(fake_root, managed_account):
    args = (fake_root, "DELETE", BASE_URL + "/managed-accounts/my_acc")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        managed_account.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = managed_account.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)



from unittest import mock

import pytest

from snowflake.core import PollingOperation
from snowflake.core.pipe import Pipe, PipeResource
from snowflake.core.pipe._generated import TagAssignment, TagReference
from snowflake.core.tag import TagValue

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
PIPE = Pipe(name="my_pipe", copy_statement="copy into my_tab")


@pytest.fixture
def pipes(schema):
    return schema.pipes


@pytest.fixture
def pipe(pipes):
    return pipes["my_pipe"]


def test_create_pipe(fake_root, pipes):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/pipes")
    kwargs = extra_params(query_params=[], body={"name": "my_pipe", "copy_statement": "copy into my_tab"})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        pipe_res = pipes.create(PIPE)
        assert isinstance(pipe_res, PipeResource)
        assert pipe_res.name == "my_pipe"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = pipes.create_async(PIPE)
        assert isinstance(op, PollingOperation)
        et_res = op.result()
        assert et_res.name == "my_pipe"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_pipe(fake_root, pipes):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/pipes")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        pipes.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = pipes.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_pipe(fake_root, pipe):
    from snowflake.core.pipe._generated.models import Pipe as PipeModel

    model = PipeModel(name="my_pipe", copy_statement="copy into my_tab")
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/pipes/my_pipe")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        pipe.fetch()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        op = pipe.fetch_async()
        assert isinstance(op, PollingOperation)
        tab = op.result()
        assert tab.to_dict() == PIPE.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_pipe(fake_root, pipe):
    args = (fake_root, "DELETE", BASE_URL + "/databases/my_db/schemas/my_schema/pipes/my_pipe")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        pipe.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = pipe.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_refresh_pipe(fake_root, pipe):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/pipes/my_pipe:refresh")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        pipe.refresh()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = pipe.refresh_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_set_tags(fake_root, pipe, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/pipes/my_pipe:set-tags")
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        pipe.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = pipe.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, pipe, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/pipes/my_pipe:unset-tags")
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        pipe.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = pipe.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, pipe):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/pipes/my_pipe:get-tags")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert pipe.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = pipe.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



from typing import Optional

import pytest

from snowflake.core.function import FunctionCollection
from snowflake.core.tag import TagResource, TagValue

from ..base_tag_tests import BaseTagTests
from ..utils import random_string
from .utils import create_service_function


pytestmark = [pytest.mark.skip_gov]


class TestFunctionTags(BaseTagTests):
    @pytest.fixture(autouse=True)
    def setup(self, temp_service_for_function, functions: FunctionCollection):
        self.functions = functions
        function_name = random_string(6, "function_for_tag_tests_")
        function_name_with_args = f"{function_name}(REAL)"
        self.function_res = create_service_function(
            function_name,
            ["REAL"],
            "REAL",
            "ep1",
            temp_service_for_function.name,
            functions,
        )
        self.function_name_with_args = function_name_with_args
        try:
            yield
        finally:
            functions[self.function_name_with_args].drop()

    @property
    def resource_level_name(self) -> str:
        return "FUNCTION"

    def set_tags(
        self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = False, resource_name: Optional[str] = None
    ):
        self.functions[self._normalize_resource_name(resource_name) or self.function_name_with_args].set_tags(
            tags, if_exists
        )

    def unset_tags(
        self, tag_resources: set[TagResource], if_exists: Optional[bool] = False, resource_name: Optional[str] = None
    ):
        self.functions[self._normalize_resource_name(resource_name) or self.function_name_with_args].unset_tags(
            tag_resources, if_exists
        )

    def get_tags(
        self, with_lineage: Optional[bool] = False, resource_name: Optional[str] = None
    ) -> dict[TagResource, TagValue]:
        return self.functions[self._normalize_resource_name(resource_name) or self.function_name_with_args].get_tags(
            with_lineage
        )

    @staticmethod
    def _normalize_resource_name(name: str | None) -> str | None:
        return f"{name}()" if name and not name.endswith(")") else name



from unittest import mock

import pytest

from snowflake.core import PollingOperation
from snowflake.core.role import ContainingScope, Role, RoleCollection, RoleResource, Securable
from snowflake.core.role._generated import TagAssignment, TagReference
from snowflake.core.tag import TagValue

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
ROLE = Role(name="my_role")
OTHER_ROLE = Securable(name="my_role")


@pytest.fixture
def roles(fake_root):
    return RoleCollection(fake_root)


@pytest.fixture
def role(roles):
    return roles["my_role"]


def test_create_role(fake_root, roles):
    args = (fake_root, "POST", BASE_URL + "/roles?createMode=errorIfExists")
    kwargs = extra_params(query_params=[("createMode", "errorIfExists")], body={"name": "my_role"})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        role_res = roles.create(ROLE)
        assert isinstance(role_res, RoleResource)
        assert role_res.name == "my_role"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = roles.create_async(ROLE)
        assert isinstance(op, PollingOperation)
        role_res = op.result()
        assert role_res.name == "my_role"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_role(fake_root, roles):
    args = (fake_root, "GET", BASE_URL + "/roles")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        it = roles.iter()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = roles.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_role(fake_root, role):
    args = (fake_root, "DELETE", BASE_URL + "/roles/my_role")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        role.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = role.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


@pytest.mark.parametrize("method, fn", [("grants", "grant_role"), ("grants:revoke", "revoke_role")])
def test_grant_revoke_role(fake_root, role, method, fn):
    args = (fake_root, "POST", BASE_URL + f"/roles/my_role/{method}")
    kwargs = extra_params(body={"securable": {"name": "my_role"}, "securable_type": "ROLE"})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        getattr(role, fn)("ROLE", OTHER_ROLE)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = getattr(role, fn + "_async")("ROLE", OTHER_ROLE)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


@pytest.mark.parametrize("method, fn", [("grants", "grant_privileges"), ("grants:revoke", "revoke_privileges")])
def test_grant_revoke_privileges(fake_root, role, method, fn):
    args = (fake_root, "POST", BASE_URL + f"/roles/my_role/{method}")
    kwargs = extra_params(body={"securable_type": "database", "privileges": ["USAGE"]})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        getattr(role, fn)(["USAGE"], "database")
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = getattr(role, fn + "_async")(["USAGE"], "database")
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


@pytest.mark.parametrize(
    "method, fn",
    [
        ("grants", "grant_privileges_on_all"),
        ("grants:revoke", "revoke_privileges_on_all"),
        ("future-grants", "grant_future_privileges"),
        ("future-grants:revoke", "revoke_future_privileges"),
    ],
)
def test_grant_revoke_future_privileges_on_all(fake_root, role, method, fn):
    args = (fake_root, "POST", BASE_URL + f"/roles/my_role/{method}")
    kwargs = extra_params(
        body={"containing_scope": {"database": "my_db"}, "securable_type": "database", "privileges": ["USAGE"]}
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        getattr(role, fn)(["USAGE"], "database", ContainingScope(database="my_db"))
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = getattr(role, fn + "_async")(["USAGE"], "database", ContainingScope(database="my_db"))
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_revoke_grant_option_for_privileges(fake_root, role):
    args = (fake_root, "POST", BASE_URL + "/roles/my_role/grants:revoke")
    kwargs = extra_params(body={"securable_type": "database", "grant_option": True, "privileges": ["USAGE"]})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        role.revoke_grant_option_for_privileges(["USAGE"], "database")
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = role.revoke_grant_option_for_privileges_async(["USAGE"], "database")
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


@pytest.mark.parametrize(
    "method, fn",
    [
        ("grants:revoke", "revoke_grant_option_for_privileges_on_all"),
        ("future-grants:revoke", "revoke_grant_option_for_future_privileges"),
    ],
)
def test_revoke_grant_option_for_future_privileges_on_all(fake_root, role, method, fn):
    args = (fake_root, "POST", BASE_URL + f"/roles/my_role/{method}")
    kwargs = extra_params(
        body={
            "containing_scope": {"database": "my_db"},
            "securable_type": "database",
            "grant_option": True,
            "privileges": ["USAGE"],
        }
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        getattr(role, fn)(["USAGE"], "database", ContainingScope(database="my_db"))
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = getattr(role, fn + "_async")(["USAGE"], "database", ContainingScope(database="my_db"))
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


@pytest.mark.parametrize(
    "method, fn",
    [
        ("grants-of", "iter_grants_of"),
        ("grants-on", "iter_grants_on"),
        ("grants", "iter_grants_to"),
        ("future-grants", "iter_future_grants_to"),
    ],
)
def test_iter_grants_to(fake_root, role, method, fn):
    args = (fake_root, "GET", BASE_URL + f"/roles/my_role/{method}")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        getattr(role, fn)()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = getattr(role, fn + "_async")()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_set_tags(fake_root, role, tag):
    args = (fake_root, "POST", BASE_URL + "/roles/my_role:set-tags")
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        role.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = role.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, role, tag):
    args = (fake_root, "POST", BASE_URL + "/roles/my_role:unset-tags")
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        role.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = role.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, role):
    args = (fake_root, "GET", BASE_URL + "/roles/my_role:get-tags")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert role.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = role.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



"""Manages Snowflake Network Policies."""

from ..network_policy._generated.models import NetworkPolicy
from ._network_policy import NetworkPolicyCollection, NetworkPolicyResource


__all__ = ["NetworkPolicy", "NetworkPolicyResource", "NetworkPolicyCollection"]



from collections.abc import Iterable, Iterator
from concurrent.futures import CancelledError, Future
from typing import Any, Callable, TypeVar


T = TypeVar("T")
U = TypeVar("U")


class PollingOperation(Future[U]):
    """Encapsulates the asynchronous execution of a callable.

    Supports all methods provided by the :class:`concurrent.futures.Future` class. :class:`PollingOperation` instances
    are not intended to be created directly.

    :param: transformer - a callable that is executed with the result of the delegated future as an argument.
        The result of the callable is accessible via the :func:`result` method.

    Examples
    ________
    >>> operation = root.databases.iter_async()

    Getting a value returned by the call, waiting indefinitely if the call hasn't completed yet:

    >>> result = operation.result()

    Getting a value returned by the call, waiting up to 60 seconds if the call hasn't completed yet:

    >>> result = operation.result(timeout=60)

    Getting whether the call is currently being executed:

    >>> is_running = operation.running()

    Cancelling the call if it has not started yet:

    >>> cancelled = operation.cancel()

    Getting an exception raised by the call, waiting indefinitely if the call hasn't completed yet:

    >>> exception = operation.exception()

    Getting an exception raised by the call, waiting up to 60 seconds if the call hasn't completed yet:

    >>> exception = operation.exception(timeout=60)

    Scheduling 100 task creation jobs and waiting for all of them to complete:

    >>> operations = [
    ...     task_collection.create_async(Task(name=f"task_{n}", definition="select 1")) for n in range(100)
    ... ]
    >>> concurrent.futures.wait(operations)
    """

    def __init__(self, delegate: Future[T], transformer: Callable[[T], U]) -> None:
        self._delegate = delegate
        self._transformer = transformer
        super().__init__()
        delegate.add_done_callback(self._callback)

    def cancel(self) -> bool:
        if not self._delegate.cancel():
            return False
        return super().cancel()

    def _callback(self, _: Any) -> None:
        try:
            result = self._delegate.result()
            self.set_result(self._transformer(result))
        except CancelledError:
            return  # do nothing
        except Exception as e:
            self.set_exception(e)


class PollingOperations:
    @staticmethod
    def empty(future: Future[T]) -> PollingOperation[None]:
        return PollingOperation(future, lambda _: None)

    @staticmethod
    def identity(future: Future[T]) -> PollingOperation[T]:
        return PollingOperation(future, lambda x: x)

    @staticmethod
    def iterator(future: Future[Iterable[T]]) -> PollingOperation[Iterator[T]]:
        return PollingOperation(future, lambda x: iter(x))



import pytest as pytest

from tests.integ.utils import random_string


@pytest.fixture
def test_database_role_name(connection):
    database_role_name = random_string(4, "test_database_grant_role_")
    with connection.cursor() as cursor:
        cursor.execute(f"create database role if not exists {database_role_name}")
        try:
            yield database_role_name
        finally:
            cursor.execute(f"drop database role if exists {database_role_name}")


@pytest.fixture
def test_share_name(connection):
    share_name = random_string(4, "test_share_")
    with connection.cursor() as cursor:
        cursor.execute(f"create share if not exists {share_name}")
        try:
            yield share_name
        finally:
            cursor.execute(f"drop share if exists {share_name}")


@pytest.fixture
def test_role_name(connection):
    role_name = random_string(4, "test_grant_role_")
    with connection.cursor() as cursor:
        cursor.execute(f"create role if not exists {role_name}")
        try:
            yield role_name
        finally:
            cursor.execute(f"drop role if exists {role_name}")


@pytest.fixture
def test_user_name(connection):
    user_name = random_string(10, "test_user_")
    with connection.cursor() as cursor:
        cursor.execute(f"create user if not exists {user_name}")
        try:
            yield user_name
        finally:
            cursor.execute(f"drop user if exists {user_name}")


@pytest.fixture
def test_table_name(connection):
    table_name = random_string(10, "test_table_")
    with connection.cursor() as cursor:
        cursor.execute(f"CREATE TABLE {table_name}(a INT)")
        try:
            yield table_name
        finally:
            cursor.execute(f"drop table if exists {table_name}")


@pytest.fixture
def test_warehouse_name(connection):
    warehouse_name = random_string(10, "test_warehouse_")
    with connection.cursor() as cursor:
        cursor.execute(f"create warehouse {warehouse_name}")
        try:
            yield warehouse_name
        finally:
            cursor.execute(f"drop warehouse if exists {warehouse_name}")


@pytest.fixture
def test_function_name(connection):
    function_name = random_string(10, "test_function_")
    with connection.cursor() as cursor:
        cursor.execute(
            f"CREATE FUNCTION {function_name}(a number, b number) RETURNS number COMMENT='multiply two numbers' AS 'a * b'"
        )
        try:
            yield function_name
        finally:
            cursor.execute(f"drop FUNCTION if exists {function_name}(number, number)")



import pytest

from snowflake.core._common import DeleteMode
from snowflake.core.grant._grant import Grant
from snowflake.core.grant._grantee import Grantees
from snowflake.core.grant._privileges import Privileges
from snowflake.core.grant._securables import Securables


@pytest.mark.use_accountadmin
@pytest.mark.usefixtures("backup_warehouse_fixture")
def test_revoke_grant(
    grants, session, schema, test_role_name, test_database_role_name, test_warehouse_name, test_function_name
):
    # grant roles

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.all_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select],
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.function(f"{test_function_name}(number, number)"),
            privileges=[Privileges.all_privileges],
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.future_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select, Privileges.insert],
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.database_role(test_database_role_name),
            securable=Securables.all_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select],
        )
    )

    # check roles have been granted
    grants_to_role = session.sql(f"SHOW GRANTS TO ROLE {test_role_name}").collect()
    grants_list = [grant for grant in grants_to_role]
    assert len(grants_list) > 0
    grants_to_database_role = session.sql(f"SHOW GRANTS TO DATABASE ROLE {test_database_role_name}").collect()
    grants_list = [grant for grant in grants_to_database_role]
    assert len(grants_list) > 0

    # revoke roles
    grants.revoke(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
        ),
        mode=DeleteMode.cascade,
    )

    grants.revoke(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
        )
    )

    grants.revoke(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.all_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select],
        )
    )

    grants.revoke(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.function(f"{test_function_name}(number, number)"),
            privileges=[Privileges.all_privileges],
        )
    )

    grants.revoke(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.future_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select, Privileges.insert],
        )
    )

    grants.revoke(
        Grant(
            grantee=Grantees.database_role(test_database_role_name),
            securable=Securables.all_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select],
        )
    )

    # check revoked roles have been removed
    grants_to_role = session.sql(f"SHOW GRANTS TO ROLE {test_role_name}").collect()
    grants_list = [grant for grant in grants_to_role]
    assert len(grants_list) == 0
    grants_to_database_role = session.sql(f"SHOW GRANTS TO DATABASE ROLE {test_database_role_name}").collect()
    grants_list = [grant for grant in grants_to_database_role]
    assert len(grants_list) == 1  # USAGE ROLE WOULD ALWAYS BE THERE


@pytest.mark.use_accountadmin
@pytest.mark.usefixtures("backup_warehouse_fixture")
def test_revoke_grant_with_grant_option(
    grants, session, schema, test_role_name, test_database_role_name, test_warehouse_name, test_function_name
):
    # grant roles

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
            grant_option=True,
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
            grant_option=True,
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.all_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select],
            grant_option=True,
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.function(f"{test_function_name}(number, number)"),
            privileges=[Privileges.all_privileges],
            grant_option=True,
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.future_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select, Privileges.insert],
            grant_option=True,
        )
    )

    grants.grant(
        Grant(
            grantee=Grantees.database_role(test_database_role_name),
            securable=Securables.all_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select],
            grant_option=True,
        )
    )

    # check roles have been granted
    grants_to_role = session.sql(f"SHOW GRANTS TO ROLE {test_role_name}").collect()
    grants_list = [grant for grant in grants_to_role]
    assert len(grants_list) > 0
    grants_to_database_role = session.sql(f"SHOW GRANTS TO DATABASE ROLE {test_database_role_name}").collect()
    grants_list = [grant for grant in grants_to_database_role]
    assert len(grants_list) > 0

    # revoke roles
    grants.revoke_grant_option(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
        ),
        mode=DeleteMode.cascade,
    )

    grants.revoke_grant_option(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.warehouse(test_warehouse_name),
            privileges=[Privileges.operate],
        )
    )

    grants.revoke_grant_option(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.all_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select],
        )
    )

    grants.revoke_grant_option(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.function(f"{test_function_name}(number, number)"),
            privileges=[Privileges.all_privileges],
        )
    )

    grants.revoke_grant_option(
        Grant(
            grantee=Grantees.role(test_role_name),
            securable=Securables.future_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select, Privileges.insert],
        )
    )

    grants.revoke_grant_option(
        Grant(
            grantee=Grantees.database_role(test_database_role_name),
            securable=Securables.all_tables(Securables.schema(schema.name)),
            privileges=[Privileges.select],
        )
    )

    # grant option removal does not remove already existing privileges
    grants_to_role = session.sql(f"SHOW GRANTS TO ROLE {test_role_name}").collect()
    grants_list = [grant for grant in grants_to_role]
    assert len(grants_list) > 0
    grants_to_database_role = session.sql(f"SHOW GRANTS TO DATABASE ROLE {test_database_role_name}").collect()
    grants_list = [grant for grant in grants_to_database_role]
    assert len(grants_list) > 0  # USAGE ROLE WOULD ALWAYS BE THERE



import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.notebook import Notebook

from ..utils import random_string
from .conftest import notebook_file


@pytest.mark.min_sf_ver("8.37.0")
def test_rename(temp_db, notebooks, notebook_stage_with_file, warehouse):
    notebook_name = random_string(10, "test_original_notebook")
    notebook_other_name = random_string(10, "test_other_notebook")

    nb = notebooks.create(
        Notebook(
            name=notebook_name,
            query_warehouse=warehouse.name,
            from_location=f"@{notebook_stage_with_file.database.name}"
            f".{notebook_stage_with_file.schema.name}."
            f"{notebook_stage_with_file.name}",
            main_file=notebook_file,
        )
    )
    try:
        nb.rename(notebook_other_name)
        assert nb.fetch().name.upper() == notebook_other_name.upper()
        with pytest.raises(NotFoundError):
            notebooks[notebook_name].fetch()
        nb.rename(nb.fetch().name, target_database=temp_db.name, target_schema="PUBLIC")
        final_def = nb.fetch()
        assert final_def.database_name.upper() == temp_db.name.upper()
        assert final_def.schema_name.upper() == "PUBLIC"
        assert final_def.name.upper() == notebook_other_name.upper()
    finally:
        nb.drop()



"""
Snowflake Sequence API.

The Snowflake Sequence API is a REST API that you can use to access, update, and perform certain actions on Sequence resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictInt, StrictStr, field_validator
from typing_extensions import Annotated


class Sequence(BaseModel):
    """A model object representing the Sequence resource.

    Constructs an object of type Sequence with the provided properties.

    Parameters
    __________
    name : str
        Name of the sequence
    start : int, optional
        First value returned by the sequence
    increment : int, optional
        Step interval of the sequence
    ordered : bool, optional
        Whether or not the values are generated in increasing or decreasing order
    comment : str, optional
        Comment for the sequence
    created_on : datetime, optional
        Date and time when the sequence was created — **Read-only:** *any user-provided value will be ignored.*
    database_name : str, optional
        Database in which the sequence is stored — **Read-only:** *any user-provided value will be ignored.*
    schema_name : str, optional
        Schema in which the sequence is stored — **Read-only:** *any user-provided value will be ignored.*
    owner : str, optional
        Role that owns the sequence — **Read-only:** *any user-provided value will be ignored.*
    owner_role_type : str, optional
        The type of role that owns the sequence — **Read-only:** *any user-provided value will be ignored.*
    """

    start: Optional[StrictInt] = None

    increment: Optional[StrictInt] = None

    ordered: Optional[StrictBool] = None

    comment: Optional[StrictStr] = None

    name: Annotated[str, Field(strict=True)]

    created_on: Optional[datetime] = None

    database_name: Optional[Annotated[str, Field(strict=True)]] = None

    schema_name: Optional[Annotated[str, Field(strict=True)]] = None

    owner: Optional[Annotated[str, Field(strict=True)]] = None

    owner_role_type: Optional[StrictStr] = None

    __properties = [
        "start",
        "increment",
        "ordered",
        "comment",
        "name",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "owner_role_type",
    ]

    @field_validator("name")
    def name_validate_regular_expression(cls, v):
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("database_name")
    def database_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("schema_name")
    def schema_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("owner")
    def owner_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Sequence:
        """Create an instance of Sequence from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "database_name",
                    "schema_name",
                    "owner",
                    "owner_role_type",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Sequence:
        """Create an instance of Sequence from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return Sequence.model_validate(obj)

        _obj = Sequence.model_validate(
            {
                "start": obj.get("start"),
                "increment": obj.get("increment"),
                "ordered": obj.get("ordered"),
                "comment": obj.get("comment"),
                "name": obj.get("name"),
                "created_on": obj.get("created_on"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "owner": obj.get("owner"),
                "owner_role_type": obj.get("owner_role_type"),
            }
        )

        return _obj


class SequenceModel:
    def __init__(
        self,
        name: str,
        # optional properties
        start: Optional[int] = None,
        increment: Optional[int] = None,
        ordered: Optional[bool] = None,
        comment: Optional[str] = None,
        created_on: Optional[datetime] = None,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        owner: Optional[str] = None,
        owner_role_type: Optional[str] = None,
    ):
        """A model object representing the Sequence resource.

        Constructs an object of type Sequence with the provided properties.

        Parameters
        __________
        name : str
            Name of the sequence
        start : int, optional
            First value returned by the sequence
        increment : int, optional
            Step interval of the sequence
        ordered : bool, optional
            Whether or not the values are generated in increasing or decreasing order
        comment : str, optional
            Comment for the sequence
        created_on : datetime, optional
            Date and time when the sequence was created.
        database_name : str, optional
            Database in which the sequence is stored
        schema_name : str, optional
            Schema in which the sequence is stored
        owner : str, optional
            Role that owns the sequence
        owner_role_type : str, optional
            The type of role that owns the sequence
        """
        self.start = start
        self.increment = increment
        self.ordered = ordered
        self.comment = comment
        self.name = name
        self.created_on = created_on
        self.database_name = database_name
        self.schema_name = schema_name
        self.owner = owner
        self.owner_role_type = owner_role_type

    __properties = [
        "start",
        "increment",
        "ordered",
        "comment",
        "name",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "owner_role_type",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Sequence(
            start=self.start,
            increment=self.increment,
            ordered=self.ordered,
            comment=self.comment,
            name=self.name,
            created_on=self.created_on,
            database_name=self.database_name,
            schema_name=self.schema_name,
            owner=self.owner,
            owner_role_type=self.owner_role_type,
        )

    @classmethod
    def _from_model(cls, model) -> SequenceModel:
        return SequenceModel(
            start=model.start,
            increment=model.increment,
            ordered=model.ordered,
            comment=model.comment,
            name=model.name,
            created_on=model.created_on,
            database_name=model.database_name,
            schema_name=model.schema_name,
            owner=model.owner,
            owner_role_type=model.owner_role_type,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a Sequence.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Sequence object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> SequenceModel:
        """Create an instance of Sequence from a dict.

        This method constructs a Sequence object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Sequence
            A Sequence object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Sequence.from_dict(obj))


Sequence._model_class = SequenceModel
